/*
 ***********************************************************************************************************************
 *
 *  Copyright (c) 2015-2025 Advanced Micro Devices, Inc. All Rights Reserved.
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in all
 *  copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 *  SOFTWARE.
 *
 **********************************************************************************************************************/

#pragma once

#include "core/hw/gfxip/gfxBlendOptimizer.h"
#include "core/hw/gfxip/gfxCmdBuffer.h"
#include "palDeque.h"

namespace Pal
{

class      PerfExperiment;
enum class ShaderType : uint32;

// Set of flags indicating which graphics states have been modified in a command buffer.
union GraphicsStateFlags
{
    struct
    {
        uint32 colorBlendState        : 1;
        uint32 depthStencilState      : 1;
        uint32 msaaState              : 1;
        uint32 quadSamplePatternState : 1;
        uint32 viewports              : 1;
        uint32 scissorRects           : 1;
        uint32 inputAssemblyState     : 1;
        uint32 triangleRasterState    : 1;
        uint32 occlusionQueryActive   : 1;
        uint32 lineStippleState       : 1;
        uint32 colorTargetView        : 1;
        uint32 depthStencilView       : 1;
        uint32 vrsRateParams          : 1;
        uint32 vrsCenterState         : 1;
        uint32 vrsImage               : 1;
        uint32 streamOutTargets       : 1;
        uint32 iaState                : 1;
        uint32 blendConstState        : 1;
        uint32 depthBiasState         : 1;
        uint32 depthBoundsState       : 1;
        uint32 pointLineRasterState   : 1;
        uint32 stencilRefMaskState    : 1;
        uint32 globalScissorState     : 1;
        uint32 clipRectsState         : 1;
        uint32 pipelineStatsQuery     : 1;
        uint32 streamoutStatsQuery    : 1;
        uint32 reserved               : 6;
    };

    uint32 u32All;
};

static_assert(sizeof(GraphicsStateFlags) == sizeof(uint32), "Bad bitfield size.");

// The Max rectangle number that is allowed for clip rects.
constexpr uint32 MaxClipRects = 4;

// The default value of clip rule which means no clip rectangles.
constexpr uint16 DefaultClipRectsRule = 0xFFFF;

// Represents the graphics state which is currently active within the command buffer.
struct GraphicsState
{
    PipelineState               pipelineState;

    DynamicGraphicsShaderInfos  dynamicGraphicsInfo; // Dynamic gfx pipeline info.
    DynamicGraphicsState        dynamicState;        // Dynamic pipeline bind state.

    BindTargetParams            bindTargets;
    // Lower MaxColorTargets bits are used. Each indicate how this slot is bound.
    // 0 indicates that it's bound to NULL, 1 means it's bound to a color target.
    uint32                      boundColorTargetMask;
    Extent2d                    targetExtent;

    BindStreamOutTargetParams   bindStreamOutTargets;

    const IColorBlendState*     pColorBlendState;
    const IDepthStencilState*   pDepthStencilState;
    const IMsaaState*           pMsaaState;

    UserDataEntries             gfxUserDataEntries;

    struct
    {
        gpusize                 indexAddr;            // GPU virtual address of the index buffer data
        uint32                  indexCount;           // Number of indices in the index buffer
        IndexType               indexType;            // Data type of the indices
    } iaState; // Input Assembly State

    InputAssemblyStateParams    inputAssemblyState;     // Current input assembler state
    BlendConstParams            blendConstState;        // (CmdSetBlendConst)
    DepthBiasParams             depthBiasState;         // (CmdSetDepthBiasState)
    DepthBoundsParams           depthBoundsState;       // (CmdSetDepthBounds)
    PointLineRasterStateParams  pointLineRasterState;   // (CmdSetPointLineRasterState)
    LineStippleStateParams      lineStippleState;       // (CmdSetLineStippleState)
    StencilRefMaskParams        stencilRefMaskState;    // (CmdSetStencilRefMasks)
    TriangleRasterStateParams   triangleRasterState;    // (CmdSetTriangleRasterState)
    ViewportParams              viewportState;          // (CmdSetViewports)
    ScissorRectParams           scissorRectState;       // (CmdSetScissorRects)
    GlobalScissorParams         globalScissorState;     // (CmdSetGlobalScissor)
    MsaaQuadSamplePattern       quadSamplePatternState; // (CmdSetQuadSamplePattern)

    VrsRateParams               vrsRateState;           // (CmdSetPerDrawVrsRate)
    VrsCenterState              vrsCenterState;         // (CmdSetVrsCenterState)
    const Image*                pVrsImage;              // (CmdBindSampleRateImage)

    uint32                      numSamplesPerPixel;     // (CmdSetQuadSamplePattern)

    uint32                      viewInstanceMask;       // (CmdSetViewInstanceMask)

    struct
    {
        uint32  enableMultiViewport    : 1;  // Is the current pipeline using viewport-array-index?
        uint32  depthClampMode         : 2;  // The current pipeline's depth clamp mode. The is of type DepthClampMode
        uint32  useCustomSamplePattern : 1;  // If use custom sample pattern instead of default sample pattern
    };

    InheritedStateParams inheritedState; // States provided to nested command buffer from primary command buffer.

    struct
    {
        uint16 clipRule;
        uint32 rectCount;
        Rect   rectList[MaxClipRects];
    } clipRectsState; // (CmdSetClipRects)

    GraphicsStateFlags   dirtyFlags;
    GraphicsStateFlags   leakFlags;      // Graphics state which a nested command buffer "leaks" back to its caller.
};

struct ValidateDrawInfo
{
    uint32       vtxIdxCount;            // Vertex or index count for the draw (depending on if the it is indexed).
    uint32       instanceCount;          // Instance count for the draw. A count of zero indicates draw-indirect.
    uint32       firstVertex;            // First vertex
    uint32       firstInstance;          // First instance
    uint32       firstIndex;             // First index
    uint32       drawIndex;              // draw index
    DispatchDims meshDispatchDims;       // Dispatch dimensions of mesh shader
    uint32       indirectDrawArgsHi;     // High part of GPU virtual address argument from the client indirect draw
    struct
    {
        uint32   useOpaque          :  1; // If draw opaque
        uint32   reserved0          :  7; // Reserved
        uint32   multiIndirectDraw  :  1; // Is multi indirect draw?
        uint32   reserved1          :  7; // Reserved
        uint32   isAdvancedIndirect :  1; // Is Execute Indirect or task+mesh draw call?
        uint32   reserved2          :  7; // Reserved
        uint32   isIndirect         :  1; // Is indirect?
        uint32   isIndexed          :  1; // Is indexed?
        uint32   reserved3          :  6; // Reserved
    };
};

// =====================================================================================================================
// Class for executing basic hardware-specific functionality common to all PM4 universal command buffers.
class UniversalCmdBuffer : public GfxCmdBuffer
{
public:
    virtual Result Begin(const CmdBufferBuildInfo& info) override;
    virtual Result End() override;
    virtual Result Reset(ICmdAllocator* pCmdAllocator, bool returnDataChunks) override;

    virtual void CmdBindPipeline(
        const PipelineBindParams& params) override;

    virtual void CmdDuplicateUserData(
        PipelineBindPoint source,
        PipelineBindPoint dest) override;

    virtual void CmdBindIndexData(
        gpusize   gpuAddr,
        uint32    indexCount,
        IndexType indexType) override;

    virtual void CmdSetViewInstanceMask(uint32 mask) override;

    virtual void CmdSetLineStippleState(
        const LineStippleStateParams& params) override;

    virtual void CmdSetPerDrawVrsRate(const VrsRateParams&  rateParams) override;

    virtual void CmdSetVrsCenterState(const VrsCenterState&  centerState) override;

    virtual void CmdBindSampleRateImage(const IImage*  pImage) override;

    // This function allows us to dump the contents of this command buffer to a file at submission time.
    virtual void DumpCmdStreamsToFile(Util::File* pFile, CmdBufDumpFormat mode) const override;
    virtual void EndCmdBufferDump(const Pal::CmdStream** ppCmdStreams, uint32 cmdStreamsNum) override;

    // Universal command buffers have two command streams: Draw Engine and a hidden ACE cmd stream.
    static constexpr uint32 NumCmdStreamsVal = 2;
    static constexpr uint32 AceStreamCount   = 1;

    // Returns the number of command streams associated with this command buffer.
    virtual uint32 NumCmdStreams() const override { return NumCmdStreamsVal; }

    virtual const CmdStream* GetCmdStream(uint32 cmdStreamIdx) const override;

    virtual const CmdStream* GetCmdStreamInSubQueue(int32 subQueueIndex) const override;

    // Universal command buffers support every type of query
    virtual bool IsQueryAllowed(QueryPoolType queryPoolType) const override { return true; }

    virtual void CmdSaveGraphicsState() override;
    virtual void CmdRestoreGraphicsStateInternal(bool trackBltActiveFlags = true) override;

    // Increments the submit-count of the command stream(s) contained in this command buffer.
    virtual void IncrementSubmitCount() override
    {
        m_pCmdStream->IncrementSubmitCount();

        if (m_pAceCmdStream != nullptr)
        {
            m_pAceCmdStream->IncrementSubmitCount();
        }
    }

    virtual uint32 GetUsedSize(CmdAllocType type) const override;

    // Current graphics state
    const GraphicsState& GetGraphicsState() const { return m_graphicsState; }

    static void SetStencilRefMasksState(
        const StencilRefMaskParams& updatedRefMaskState,
        StencilRefMaskParams*       pStencilRefMaskState);

    // Used to initialize boundColorTargetMask. Null color target is bound only when the slot was not NULL and being
    // bound to NULL. Set all 1s so NULL color targets will be bound when BuildNullColorTargets() is called first time.
    static constexpr uint32 NoNullColorTargetMask = ((1 << MaxColorTargets) - 1);

    virtual void CmdBindPipelineWithOverrides(
        const PipelineBindParams& params,
        SwizzledFormat            swizzledFormat,
        uint32                    targetIndex) override;

    void CmdCloneImageData(const IImage& srcImage, const IImage& dstImage);

protected:
    UniversalCmdBuffer(
        const GfxDevice&           device,
        const CmdBufferCreateInfo& createInfo,
        const GfxBarrierMgr&       barrierMgr,
        GfxCmdStream*              pDeCmdStream,
        GfxCmdStream*              pAceCmdStream,
        bool                       blendOptEnable,
        bool                       useUpdateUserData);

    virtual ~UniversalCmdBuffer() {}

    virtual Result BeginCommandStreams(CmdStreamBeginFlags cmdStreamFlags, bool doReset) override;

    virtual void ResetState() override;

    void LeakNestedCmdBufferState(
        const UniversalCmdBuffer& cmdBuffer);

    static void PAL_STDCALL CmdSetUserDataGfxFiltered(
        ICmdBuffer*   pCmdBuffer,
        uint32        firstEntry,
        uint32        entryCount,
        const uint32* pEntryValues);

    static void PAL_STDCALL CmdUpdateUserDataGfx(
        ICmdBuffer*   pCmdBuffer,
        uint32        firstEntry,
        uint32        entryCount,
        const uint32* pEntryValues);

    bool IsAnyGfxUserDataDirty() const;

    virtual void SetGraphicsState(const GraphicsState& newGraphicsState);
    virtual void SetGraphicsState(const GraphicsState& newGraphicsState,
                                  GraphicsStateFlags   setGraphicsStateFlags,
                                  PipelineStateFlags   setPipelineStateFlags);

    GraphicsState  m_graphicsState;        // Currently bound graphics command buffer state.
    GraphicsState  m_graphicsRestoreState; // State pushed by the previous call to CmdSaveGraphicsState.

    GfxBlendOptimizer::BlendOpts  m_blendOpts[MaxColorTargets]; // Current blend optimization state

    virtual uint32* WriteNops(uint32* pCmdSpace, uint32 numDwords) const override
        { return pCmdSpace + m_pCmdStream->BuildNop(numDwords, pCmdSpace); }

    // Late-initialized ACE command buffer stream.
    // Ace command stream is used for ganged submit of compute workloads (task shader workloads)
    // after which graphics workloads will be submitted on the DE command stream.
    GfxCmdStream* m_pAceCmdStream;

    TessDistributionFactors m_tessDistributionFactors;

    uint8 m_contextStatesPerBin;
    uint8 m_persistentStatesPerBin;

private:
    const bool m_blendOptEnable;

    PAL_DISALLOW_COPY_AND_ASSIGN(UniversalCmdBuffer);
    PAL_DISALLOW_DEFAULT_CTOR(UniversalCmdBuffer);
};

} // Pal
